#!/usr/bin/env perl

# Copyright 2010 Andrew Dodd & Murdoch J Gabbay
# andrewdodd13@gmail.com

# This file is part of Publications Map.

# Publications Map is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Publications Map is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See  the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Publications Map.  If not, see <http://www.gnu.org/licenses/>.

# Include some helper files
use LWP::Simple;
use XML::Simple;
use Encode;

# Include fetch-functions
require 'fetch-functions.pl';

# Get the page URL from the user
sub get_page_url() {
	print "\nPlease enter the location of your DBLP page, for example:\n",
		"http://www.informatik.uni-trier.de/~ley/db/indices/a-tree/g/Gabbay:Murdoch_James.html\n",
		"> ";
		
	chomp($res = <STDIN>);
	
	# Grab the identifier from the HTTP URL
	$res =~ /\/(.\/[^:]+:[^\.]+)\.html/;
	
	return $1;	
}

# Print messages straight away, rather than buffering them
$| = 1;

print "Publications Map: Fetch DBLP Data Script\n",
	"Copyright 2010 Andrew Dodd & Murdoch J Gabbay\n",
	"This program comes with ABSOLUTELY NO WARRANTY.  ",
	"This is free software, and you are welcome to redistribute it under certain conditions, please ",
	"read the included COPYING file for details.\n";
	
#####
# Argument mangling
#
# Check whether the user provided two correct command line options, if they did
# not then usage() if they gave incorrect ones, or prompt them to enter options
# manually.
#####

get_args('dblp');

# Create a Simple XML object
my $xp = new XML::Simple;
my $existing;

my $page_url;

# Get the source data file location
if($ARGV[0] eq "--addnew" or $ARGV[0] eq "--update") {
	$existing = $xp->XMLin("data.xml");
	$page_url = $existing->{'settings'}->{'dblpUrl'};
	
	# If URL is not set in XML
	if(not defined($page_url) or $page_url eq "" or ref($page_url) eq "HASH") {
		$page_url = get_page_url();
	}
}
elsif($#ARGV == 0) {	# One argument
	$page_url = get_page_url();
} 
else {
	$ARGV[1] =~ /\/(.\/[^:]+:[^\.]+)\.html/;
	$page_url = $1;
}

my $file_name = get_data_file();

#####
# Fetch the DBLP page
#
# Use $ARGV (which should now be set to correctly) to fetch the user's DBLP XML
# export page.
#####
print "Fetching DBLP page... ";

my $page = get("http://dblp.uni-trier.de/rec/pers/$page_url/xk");

if(substr($page, 0, 21) ne '<?xml version="1.0"?>')
{
	print "\n\nERROR: Invalid response returned by server.  Is your DBLP URL correct?\n";
	exit(1);
}

my $xml_content = $xp->XMLin($page);
print "Done.\n";

die "ERROR: Failed to parse DBLP entry.\n" unless defined $xml_content;

#####
# Fetch each record in the DBLP page
#####
print "Fetching records...\r";
my @nodes;
my $counter = 0;

for my $rec (@{$xml_content->{'dblpkey'}})
{
	next if(defined($rec->{'type'}) and $rec->{'type'} eq "person record");

	print pack("A80", "Fetching: $rec..."), "\r";

	my $current_node = { };
	my $xmlfile = Encode::decode('iso-8859-1', get("http://dblp.uni-trier.de/rec/bibtex/$rec.xml"));
	my $doc = $xp->XMLin($xmlfile);

	while(my ($type, $data) = each(%$doc))
	{
		if(substr($data->{'title'}, -1, 1) eq ".") {
			$current_node->{'title'} = substr($data->{'title'}, 0, -1);
		} else {
			$current_node->{'title'} = $data->{'title'};
		}
	
		if(ref($data->{'author'}) eq "ARRAY") {
			$current_node->{'author'} = join(' and ', @{$data->{'author'}});
		} else {
			$current_node->{'author'} = $data->{'author'};
		}		

		# In book and in collection are pretty much the same thing
		$type = "inbook" if($type eq "incollection");
		
		$current_node->{'type'} = $type;

		$current_node->{'doi'} = $data->{'ee'};
		$current_node->{'year'} = $data->{'year'};
		$current_node->{'dblp'} = $rec;

		# Just put as much information as possible into the ref field
		if($current_node->{'type'} eq "article") {
			$current_node->{'ref'} = $data->{'journal'};
			$current_node->{'ref'} .= " Volume " . $data->{'volume'} . "," if (defined($data->{'volume'}));
			$current_node->{'ref'} .= " Number " . $data->{'number'} . "," if (defined($data->{'number'}));
			$current_node->{'ref'} .= " " . $data->{'year'} . ",";
			$current_node->{'ref'} .= " Pages " . $data->{'pages'} . "," if (defined($data->{'pages'}));
			$current_node->{'ref'} = substr($current_node->{'ref'}, 0, -1);
		} elsif ($current_node->{'type'} eq "inproceedings") {
			$current_node->{'type'} = "inproceedings-c";	# map to conference
			$current_node->{'ref'} = $data->{'booktitle'} . " (" . $data->{'year'} . ") Pages " . $data->{'pages'};
		}

		$counter++;

		push(@nodes, $current_node);
	}
}

print pack("A80", "Fetching Node Details: Done."), "\n";

#####
# Parse Data
#
# All data has now been fetched from the server, we can get to work parsing it
# based on what the user wants.
#####

#####
# Create Option
#####
if($ARGV[0] eq "--create")
{
	print "Printing output... ";

	# Print a default 100x100 canvas with default bibtex types	
	my %settings = (
		width => 100,
		height => 100,
		cameraX => 0,
		cameraY => 0,
		cameraZ => 0.85,
		bibtexLocation => "",
		webBibtexLocation => "",
		demoMode => "false",
		dblpUrl => $page_url,
		dataVersion => 1.1
	);
	
	my @types = (
		{ listname => "Article", tooltip => "An article from a journal or magazine", offcolor => 15335365, oncolor => 13432202, content => "article" },
		{ listname => "Book", tooltip => "A book with an explicit publisher", offcolor => 16764855, oncolor => 16759705, content => "book" },
		{ listname => "Booklet", tooltip => "A work that is printed and bound, but without a named publisher or sponsoring institution",
			offcolor => 10934149, oncolor => 6159370, content => "booklet" },
		{ listname => "Book Chapter", tooltip => "A part of a book, usually untitled", offcolor => 16768409, oncolor => 16770223, content => "inbook" },
		{ listname => "In Collection", tooltip => "A part of a book having its own title", offcolor => 16756751, oncolor => 16763921, content => "incollection" },
		{ listname => "In Proceedings (Conference)", tooltip => "An article in a conference proceedings", offcolor => 12974079, oncolor => 9563386,
			content => "inproceedings-c" },
		{ listname => "In Proceedings (Short Abstract)", tooltip => "A short abstract", offcolor => 16777130, oncolor => 14540168, content => "inproceedings-s" },
		{ listname => "In Proceedings (Workshop)", tooltip => "An article in a workshop proceedings", offcolor => 13872889, oncolor => 12881405,
			content => "inproceedings-w" },
		{ listname => "Manual", tooltip => "Technical documentation", offcolor => 14741230, oncolor => 11849165, content => "manual" },
		{ listname => "Masters Thesis", tooltip => "A Master\'s thesis", offcolor => 16746734, oncolor => 16729309, content => "mastersthesis" },
		{ listname => "Miscellaneous", tooltip => "A paper which does not fit anywhere else", offcolor => 16777215, oncolor => 14540253, content => "misc" },
		{ listname => "PhD Thesis", tooltip => "A PhD. thesis", offcolor => 12541951, oncolor => 10170623, content => "phdthesis" },
		{ listname => "Proceedings", tooltip => "The proceedings of a conference", offcolor => 9364974, oncolor => 45806, content => "proceedings" },
		{ listname => "Tech Report", tooltip => "A report published by a school or other institution, usually numbered within a series", 
			offcolor => 13948116, oncolor => 11382189, content => "techreport" },
		{ listname => "Unpublished", tooltip => "A document having an author and title, but not formally published", offcolor => 16777215, 
			oncolor => 14540253, content => "unpublished" }
	);
	
	my $line_length = int($counter ** 0.5 + 0.5);
	my $line_pos = 0;
	my $line = 0;
	
	my @tidynodes;

	for my $current_node (@nodes)
	{
		# Convert any hashes into strings
		while (my ($k, $v) = each(%$current_node))
		{
			if(ref($v) eq "HASH") {
				$current_node->{$k} = join('', @{$v->{'content'}});
			}
		}
		
		# Hash up nodes here
		$current_node->{'id'} = $line * $line_length + $line_pos + 1;
		$current_node->{'nodex'} = $line_pos * 5 + 2;
		$current_node->{'nodey'} = $line * 2 + 2;
		$current_node->{'boxWidth'} = 4;
		
		push(@tidynodes, $current_node);

		$line_pos++;
		if($line_pos >= $line_length)
		{
			$line_pos = 0;
			$line++;
		}
	}
	
	write_xml($file_name, \%settings, \@types, \@tidynodes, [ ]);

	print "Done.\n";

	print "\nIn future, use the --addnew switch to add any new publications from DBLP.  ",
		"You may also add your own nodes with the editor.\n\n";
}
#####
# Merge Options
#####
else
{
	# Load in the existing data.xml
	# Generate node objects for each 
	print "Comparing Existing Nodes...\r";

	# Do a diff
	my $existing_nodes = $existing->{'node'};
	my $highest_id = -1;

	my %new_nodes = ( );
	my $conflict = 0;
	my $synced_node_count = 0;

	# Track conflict options from the user
	my %conflict_opts = ( );

	for my $current_node (@nodes)
	{
		print pack("A80", "Comparing: " . $current_node->{'dblp'}), "\r";
		# Convert any hashes into a string
		while (my ($k, $v) = each(%$current_node))
		{
			if(ref($v) eq "HASH") {
				$current_node->{$k} = join('', @{$v->{'content'}});
			}
		}

		# Did we find this node in data.xml?
		my $already_exists = 0;

		# Search existing nodes for this node
		while (my ($key, $val) = each(%{$existing->{node}}))
		{
			$highest_id = $key if($key > $highest_id);

			# compare dblp index
			next if(!defined($val->{'dblpUrl'}) || !defined($current_node->{'dblp'}));

			if($val->{'dblpUrl'} eq $current_node->{'dblp'}) {
				if($ARGV[0] eq "--update") {
					$conflicts = get_conflicts($val, $current_node);
					if(scalar @$conflicts != 0)
					{
						if($conflict == 0) {
							print pack("A79", "----------"), "\n",
								"Conflict resolution options: \n",
								"----------\n",
								"(t) Accept Theirs:       Use DBLP's version of this field.\n",
								"(k) Keep Mine (default): Keep your version of this field.\n",
								"(a) Accept All Theirs:   Use DBLP's version of this field for all following nodes.\n",
								"(d) Keep All Mine:       Keep your version of this field for all following nodes.\n",
								"(n) Accept Their Nodes:  Use DBLP's data for all conflicts in following nodes.\n",
								"(m) Keep My Nodes:       Keep your data for all conflicts in following nodes.\n",
								"----------\n\n";
							}
						$conflict = 1;
						conflict_resolve($val, $current_node, $conflicts, \%conflict_opts);
					}
				}
				$already_exists = 1;
				$synced_node_count++;
			}
		}

		# Add to the list of new nodes
		if ($already_exists == 0)
		{
			my %new_node = (
				'boxWidth' => '4',
				'title' => $current_node->{'title'},
				'author' => $current_node->{'author'},
				'type' => $current_node->{'type'},
				'year' => $current_node->{'year'},
				'doi' => $current_node->{'doi'},
				'ref' => $current_node->{'ref'},
				'dblpUrl' => $current_node->{'dblp'},
			);
			$new_nodes{++$highest_id} = \%new_node;

			print pack("A79", "New Node: " . $new_node{'title'}), "\n";
		}
	}

	print "Comparing Existing Nodes: Done.\n",
		scalar keys %new_nodes, " new node(s), $synced_node_count already exist.\n\n";

	if(scalar keys %new_nodes == 0 and $conflict == 0) {
		print "Nothing to do, exiting.\n\n";
		exit(0);
	}

	my $line_length = int((scalar keys %new_nodes) ** 0.5 + 0.5);
	my $line_pos = 0;
	my $line = 0;

	# Assign a position to each node
	while(my ($key, $val) = each(%new_nodes))
	{
		# +2 = margin of 2 squares
		$new_nodes{$key}{'nodex'} = $line_pos * 5 + 2;
		$new_nodes{$key}{'nodey'} = $line * 2 + 2;
		$line_pos++;
		if($line_pos >= $line_length)
		{
			$line_pos = 0;
			$line++;
		}
	}

	print "Printing output... ";

	# Print all the nodes into the XML file...
	my $merged_nodes = { %{$existing->{'node'}}, %new_nodes};
	my @xml_nodes;
	while(my ($key, $value) = each(%{$merged_nodes}))
	{
		my $xml_node = $value;
		$xml_node->{'id'} = $key;
		push(@xml_nodes, $xml_node);
	}
	
	write_xml($file_name, $existing->{'settings'}, $existing->{'types'}->{'type'}, \@xml_nodes, $existing->{'line'});

	print "Done.\n";
}

print "\nScript complete.\n\n";