#!/usr/bin/env perl

# Copyright 2010 Andrew Dodd & Murdoch J Gabbay
# andrewdodd13@gmail.com

# This file is part of Publications Map.

# Publications Map is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Publications Map is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See  the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Publications Map.  If not, see <http://www.gnu.org/licenses/>.

use constant DATA_VERSION => 1.1;

use File::Copy;

# Return a neat XML tag
sub xmlorbust {
	my ($arg, $val) = @_;
	if(!defined($val) or ref($val) eq "HASH" or $val eq "") {
		return "<$arg />";
	} else {
		$val =~ s/&/&amp;/g;
		return "<$arg>$val</$arg>";
	}
}

# Return the conflicts between two nodes
sub get_conflicts {
	my($local, $remote) = @_;
	my @conflicts;

	for my $tag ('title', 'author', 'type', 'doi', 'ref', 'year') {
		$local->{$tag} = "" if (!defined($local->{$tag}) or ref($local->{$tag}) eq "HASH");
		$remote->{$tag} = "" if (!defined($remote->{$tag}) or ref($local->{$tag}) eq "HASH");
		push (@conflicts, $tag) if($local->{$tag} ne $remote->{$tag});
	}

	return \@conflicts;
}

# Compare two nodes
sub conflict_resolve {
	my ($local, $remote, $conflicts, $opts) = @_;

	print "Conflict [", $remote->{'dblp'} , "] - ", $remote->{'title'} ,"\n";

	for my $conflict (@$conflicts) {
		if(defined $opts->{"ALLOPTS"}) {
			if($opts->{"ALLOPTS"} eq "n") {
				print "Auto-accepting $conflict\n";
				$local->{$conflict} = $remote->{$conflict};
				next;
			} elsif ($opts->{"ALLOPTS"} eq "m") {
				print "Auto-denying $conflict\n";
				next;
			}
		}
		if(defined $opts->{$conflict}) {
			# User chose to accept all
			if($opts->{$conflict} eq "a") {
				print "Auto-accepting $conflict\n";
				$local->{$conflict} = $remote->{$conflict};
				next;
			# User chose to decline all
			} elsif($opts->{$conflict} eq "d") {
				print "Auto-denying $conflict\n";
				next;
			}
		}

		print "-- [$conflict] -- \n",
			"- [Yours ] ", $local->{$conflict}, "\n",
			"- [Theirs] ", $remote->{$conflict}, "\n",
			"Resolution: [t/K/a/d/n/m]?\n> ";

		# Grab the first character only
		chomp (my $res = <STDIN>);
		$res = lc(substr($res, 0, 1));

		if($res eq "a") {
			$local->{$conflict} = $remote->{$conflict};
			$opts->{$conflict} = "a";
		} elsif ($res eq "d") {
			$opts->{$conflict} = "d";
		} elsif ($res eq "t") {
			$local->{$conflict} = $remote->{$conflict};
		} elsif ($res eq "n") {
			$local->{$conflict} = $remote->{$conflict};
			$opts->{"ALLOPTS"} = "n";
		} elsif ($res eq "m") {
			$opts->{"ALLOPTS"} = "m";
		}
	}

	print "\n";
}

# Open $filename and write out XML 
sub write_xml {
	my ($filename, $settings, $types, $nodes, $lines) = @_;
	
	move($filename, $filename . ".old");
	
	open (OUTPUT, ">:utf8", $filename);

	# Buffer file writing
	$| = 0;

	print OUTPUT '<?xml version="1.0" encoding="UTF-8"?>', "\n",
		'<data xsi:noNamespaceSchemaLocation="DataSchema.xml" ', 
		'xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">', "\n";	
		
	print OUTPUT '  <settings>',
		"\n    ", xmlorbust('width', $settings->{'width'}),
		"\n    ", xmlorbust('height', $settings->{'height'}),
		"\n    ", xmlorbust('cameraX', $settings->{'cameraX'}),
		"\n    ", xmlorbust('cameraY', $settings->{'cameraY'}),
		"\n    ", xmlorbust('cameraZ', $settings->{'cameraZ'}),
		"\n    ", xmlorbust('bibtexLocation', $settings->{'bibtexLocation'}),
		"\n    ", xmlorbust('webBibtexLocation', $settings->{'webBibtexLocation'}),
		"\n    ", xmlorbust('demoMode', $settings->{'demoMode'}),
		"\n    ", xmlorbust('dblpUrl', $settings->{'dblpUrl'}),
		"\n    ", xmlorbust('dataVersion', DATA_VERSION),
		"\n  </settings>\n";
		
	print OUTPUT '  <types>';

	for my $type (@$types)
	{
		print OUTPUT "\n", '    <type listname="', $type->{'listname'}, '" tooltip="', $type->{'tooltip'},
			'" offcolor="', $type->{'offcolor'}, '" oncolor="', $type->{'oncolor'}, '">', $type->{'content'}, 
			'</type>';
	}
	print OUTPUT "\n  </types>\n";

	for my $node (@$nodes)
	{
		print OUTPUT "  <node id=\"", $node->{'id'}, "\">\n",
			"    ", xmlorbust('nodex', $node->{'nodex'}), "\n",
			"    ", xmlorbust('nodey', $node->{'nodey'}), "\n",
			"    ", xmlorbust('boxWidth', $node->{'boxWidth'}), "\n",
			"    ", xmlorbust('title', $node->{'title'}), "\n",
			"    ", xmlorbust('author', $node->{'author'}), "\n",
			"    ", xmlorbust('type', $node->{'type'}), "\n",
			"    ", xmlorbust('year', $node->{'year'}), "\n",
			"    ", xmlorbust('month', $node->{'month'}), "\n",
			"    ", xmlorbust('file', $node->{'file'}), "\n",
			"    ", xmlorbust('doi', $node->{'doi'}), "\n",
			"    ", xmlorbust('bibtex', $node->{'bibtex'}), "\n",
			"    ", xmlorbust('tags', $node->{'tags'}), "\n",
			"    ", xmlorbust('notes', $node->{'nodes'}), "\n",
			"    ", xmlorbust('ref', $node->{'ref'}), "\n",
			"    ", xmlorbust('dblpUrl', $node->{'dblp'}), "\n",
			"    ", xmlorbust('visibleWeb', $node->{'visibleWeb'}), "\n",
			"    ", xmlorbust('visibleViewer', $node->{'visibleViewer'}), "\n",			
			"  </node>\n";	
	}
	
	# If lines is not null
	if(ref($lines) eq "HASH")
	{
		while(my ($id, $val) = each(%{$lines}))
		{
			print OUTPUT '  <line id="', $id, '">',
				"\n    ", xmlorbust('linestart', $val->{'linestart'}),
				"\n    ", xmlorbust('lineend', $val->{'lineend'}),
				"\n    ", xmlorbust('size', $val->{'size'}),
				"\n    ", xmlorbust('color', $val->{'color'}),
				"\n    ", xmlorbust('info', $val->{'info'}),
				"\n    ", xmlorbust('percentage', $val->{'percentage'}),
				"\n  </line>\n";
		}
	}
	
	print OUTPUT '</data>';

	close (OUTPUT);
}

# Print usage instructions then exit
sub usage {
	my ($datatype) = @_;
	print "Usage:\n",
		$0, " [args]\n\n",
		"--create <$datatype source> [output filename]\n",
		"  Create a blank data file with the $datatype nodes\n",
		"--addnew <$datatype source> <existing data file>\n",
		"  Use existing data file and add new $datatype nodes into it\n",
		"--update <$datatype source> <existing data file>\n",
		"  As addnew but check $datatype nodes in data file against corresponding entry in $datatype data for differences\n\n",
		"Run with no arguments for interactive mode.\n";
	exit(1);
}

#####
# Argument mangling
#
# Check whether the user provided two correct command line options, if they did
# not then usage() if they gave incorrect ones, or prompt them to enter options
# manually.
#####
sub get_args {
	my ($datatype) = @_;
	if($#ARGV > 2) {
		# Too many args
		usage($datatype);
	} elsif($#ARGV >= 0) {
		# Check the first argument is an option
		usage($datatype) if($ARGV[0] ne "--create" and $ARGV[0] ne "--update" and $ARGV[0] ne "--addnew");
	} else {
		# Prompt user for an action
		print "\nPlease choose an action: \n",
			"1 - Create a blank data file with $datatype nodes\n",
			"2 - Add new nodes from $datatype into existing Publication Map data file\n",
			"3 - As 2, also check existing $datatype nodes in data file against $datatype file for changes\n",
			"> ";
		
		chomp(my $res = <STDIN>);
		while($res ne "1" and $res ne "2" and $res ne "3")
		{
			print "Invalid choice, please enter 1, 2 or 3.\n> ";
			chomp($res = <STDIN>);
		}

		my @opts = ("--create", "--addnew", "--update");
		$ARGV[0] = $opts[$res - 1];
	}
}

sub get_data_file {
	return($#ARGV == 2 ? $ARGV[2] : "data.xml");
}

return true;